from copy import deepcopy
from pathlib import Path
import re

from openpecha.formatters.formatter import BaseFormatter
from openpecha.formatters.format import *

class HFMLFormatter(BaseFormatter):
    '''
    OpenPecha Formatter for digitized wooden-printed Pecha based on annotation scheme from Esukhia.
    '''

    def __init__(self, output_path='./output', is_book=False):
        super().__init__(output_path=output_path)
        self.is_book = is_book
        self.base_text = ''
        self.vol_walker = 0
        self.pecha_title = []
        self.poti_title = []
        self.chapter_title = []
        self.topic_id = [] # made class variable as it needs to update cross poti
        self.current_topic_id = [] # made class variable as it needs to update cross poti
        self.sub_topic = [] # made class variable as it needs to update cross poti
        self.page = []
        self.error_id = []
        self.abs_er_id = []
        self.notes_id = []
        self.sub_topic_Id = [] # made class variable as it needs to update cross poti
        self.topic_info = []
        self.sub_topic_info = []
        self.cur_sub = []
        self.author_pattern = []
        self.citation_pattern = []
        self.sabche_pattern = []
        self.tsawa_pattern = []
        self.yigchung_pattern = []


    def text_preprocess(self, text):
        if text[0] == '\ufeff':
            text = text[1:]
        
        if not self.is_book:
            return text

        p = r'\[p\]'
        lines = text.splitlines()
        result_text = ''
        para = False
        for line in (lines):
            if re.search(p,line):
                if para:
                    para = False
                    result_text += '\n'
                else:
                    para = True    
            elif re.search(p,line) == None:
                if para:
                    result_text += line
                else:
                    result_text += line+'\n'
        return result_text


    def get_input(self, input_path):
        fns = list(input_path.iterdir())
        fns_len = len(fns)
        for fn in sorted(fns):
            yield self.text_preprocess(fn.read_text()), fn.name, fns_len


    def format_layer(self, layers):
        cross_vol_anns = [layers['topic'], layers['sub_topic']]
        non_cross_vol_anns = [layers['page'], layers['correction'], layers['peydurma'], layers['error_candidate']]
        anns = {'cross_vol': cross_vol_anns, 'non_cross_vol': non_cross_vol_anns}
        for ann in anns:
            if ann == 'non_cross_vol':
                for i, (pecha_pg, pecha_correction, pecha_peydurma, pecha_error) in enumerate(zip(*anns[ann])):
                    base_id = f'v{i+1:03}'
                    # Page annotation
                    Pagination = deepcopy(Layer)
                    Pagination['id'] = self.get_unique_id()
                    Pagination['annotation_type'] = 'pagination'
                    Pagination['revision'] = f'{1:05}'
                    for start, end, pg_info, index in pecha_pg:
                        page = deepcopy(Page)
                        page['id'] = self.get_unique_id()
                        page['span']['start'] = start
                        page['span']['end'] = end
                        page['page_index'] = index
                        page['page_info'] = pg_info
                        Pagination['annotations'].append(page)

                    # Correction annotation
                    Correction_layer = deepcopy(Layer)
                    Correction_layer['id'] = self.get_unique_id()
                    Correction_layer['annotation_type'] = 'correction'
                    Correction_layer['revision'] = f'{1:05}'
                    for start, end, sug in pecha_correction:
                        correction = deepcopy(Correction)
                        correction['id'] = self.get_unique_id()
                        correction['span']['start'] = start
                        correction['span']['end'] = end
                        correction['correction'] = sug
                        Correction_layer['annotations'].append(correction)

                    # Error_candidate annotation
                    Error_layer = deepcopy(Layer)
                    Error_layer['id'] = self.get_unique_id()
                    Error_layer['annotation_type'] = 'error_candidate'
                    Error_layer['revision'] = f'{1:05}'
                    for start, end in pecha_error:
                        error = deepcopy(ErrorCandidate)
                        error['id'] = self.get_unique_id()
                        error['span']['start'] = start
                        error['span']['end'] = end
                        Error_layer['annotations'].append(error)

                    # Peydurma annotation
                    Peydurma_layer = deepcopy(Layer)
                    Peydurma_layer['id'] = self.get_unique_id()
                    Peydurma_layer['annotation_type'] = 'note_marker'
                    Peydurma_layer['revision'] = f'{1:05}'
                    for pey in pecha_peydurma:
                        peydurma = deepcopy(Peydurma)
                        peydurma['id'] = self.get_unique_id()
                        peydurma['span']['start'] = pey
                        peydurma['span']['end'] = pey
                        Peydurma_layer['annotations'].append(peydurma)


                    result = {
                        'pagination': Pagination,
                        'correction': Correction_layer,
                        'peydurma': Peydurma_layer,
                        'error_candidate': Error_layer
                    }

                    yield result, base_id
            else:
                Index_layer = deepcopy(Layer)
                Index_layer['id'] = self.get_unique_id()
                Index_layer['annotation_type'] = 'index'
                Index_layer['revision'] = f'{1:05}'
                # loop over each topic
                for topic, sub_topic in zip(*anns[ann]):
                    Topic = deepcopy(Text)
                    Topic['id'] = self.get_unique_id()

                    # loop over each sub_topic
                    for corss_sub_topic in sub_topic:
                        sub_text = deepcopy(SubText)
                        sub_text['id'] = self.get_unique_id()
                        for start, end, vol_id, work in corss_sub_topic:
                            sub_text['work'] = work
                            cross_vol_span = deepcopy(CrossVolSpan)
                            cross_vol_span['vol'] = f'base/v{vol_id:03}'
                            cross_vol_span['span']['start'] = start
                            cross_vol_span['span']['end'] = end
                            sub_text['span'].append(cross_vol_span)

                        if corss_sub_topic:
                            Topic['parts'].append(sub_text)

                    for start, end, vol_id, work in topic:
                        Topic['work'] = work
                        cross_vol_span = deepcopy(CrossVolSpan)
                        cross_vol_span['vol'] = f'base/v{vol_id:03}'
                        cross_vol_span['span']['start'] = start
                        cross_vol_span['span']['end'] = end
                        Topic['span'].append(cross_vol_span)

                    Index_layer['annotations'].append(Topic)

                result = {
                    'index': Index_layer
                }

                yield result, None


    def total_pattern(self, pat_list, annotated_line):
        """ It calculates the length of all the annotation detected in a line.

        Args:
            pat_list (dict): It contains all the annotation's regex pattern as value and name of annotation as key.
            annotated_line (str): It contains the annotated line to be process.
        
        Return:
            total_length (int): It accumulates as annotations are detected in the line.
        """
        total_length = 0 # total length of annotation detected in a line
        for pattern in ['line_pattern', 'topic_pattern', 'sub_topic_pattern']:
            if re.search(pat_list[pattern], annotated_line):
                match_list = re.finditer(pat_list[pattern], annotated_line) # list of match object of given pattern in line
                for match in match_list:
                    total_length = total_length + len(match[0])

        if re.search(pat_list['error_pattern'], annotated_line):
            errors = re.finditer(pat_list['error_pattern'], annotated_line) # list of match object of error pattern in line
            for error in errors:
                error_part = error[0].split(',')[0][1:]
                total_length = total_length + (len(error[0])-len(error_part))
            
        if re.search(pat_list['abs_er_pattern'], annotated_line):
            abs_ers = re.findall(pat_list['abs_er_pattern'], annotated_line) # list of match of abs_er pattern in line
            total_length = total_length + 2 * len(abs_ers)

        if re.search(pat_list['note_pattern'], annotated_line):
            note_patterns = re.findall(pat_list['note_pattern'], annotated_line) # list of match of note pattern in line
            total_length = total_length + len(note_patterns)

        for pattern in ['author_pattern', 'pecha_title_pattern', 'poti_title_pattern', 'chapter_title_pattern']:
            title_pattern = re.search(pat_list[pattern], annotated_line)
            if title_pattern:
                total_length += 4
        
        for pattern in ['start_cit_pattern', 'end_cit_pattern']:
            cit_patterns = re.findall(pat_list[pattern], annotated_line) # list of match of citation pattern in line
            total_length = total_length + 2 * len(cit_patterns)

        for pattern in ['start_sabche_pattern', 'end_sabche_pattern']:
            sabche_patterns = re.findall(pat_list[pattern], annotated_line) # list of match of sabche pattern in line
            total_length = total_length + 2 * len(sabche_patterns)
        
        for pattern in ['start_tsawa_pattern', 'end_tsawa_pattern']:
            tsawa_patterns = re.findall(pat_list[pattern], annotated_line) # list of match of tsawa pattern in line
            total_length = total_length + 2 * len(tsawa_patterns)
        
        for pattern in ['start_yigchung_pattern', 'end_yigchung_pattern']:
            yigchung_patterns = re.findall(pat_list[pattern], annotated_line) # list of match of yigchung pattern in line
            total_length = total_length + 2 * len(yigchung_patterns)

        return total_length

    def merge(self, start_list, end_list):
        """ It merges two list.

        The starting  and ending of annotation(citaion,yigchung,sabche and tsawa) are stored in two list.
        Merging these two list will generate a list in which both starting and ending of an annotation together in a tuple.
        It is applicable only if the annotaions are not cross volume.

        Args:
            start_list (list): It contains index of where starting annotations(citaion,yigchung,sabche and tsawa) are detected.
            end_list (list): It contains index of where ending annotations(citaion,yigchung,sabche and tsawa) are detected.
        
        Return:
            result (list): It contains tuples where starting and ending of an annotation(citaion,yigchung,sabche and tsawa) is combined.
        """
        walker = 0
        result = []
        while walker < len(end_list) and walker < len(start_list):
            result.append((start_list[walker], end_list[walker]))
            walker += 1
        return result

    def search_before(self, ann, pat_list, line): 
        """ It calculates the length of annotation detected in a given line before a given annotation.

        Args:
            ann (match object): It is a match object of the annotation of which we want to calculate
                                the length of any annotation detected before it.
            pat_list (dict): It contains all the annotation's regex pattern as value and name of annotation as key.
            line (str): It contains the line in which we wants to calculate the legth of annotation found before the given annotation.

        Return:
            length_before (int): It accumalates as we detect annotation which is before the given annotation and
                                finally gives the total length of annotation caught before the given annotation in the given line.
        """
        length_before = 0
        for pp in ['line_pattern','topic_pattern','sub_topic_pattern']:
            if re.search(pat_list[pp], line):
                match_list = re.finditer(pat_list[pp], line) # list of match object of given pattern in line
                for match in match_list:
                    if ann.start() > match.start():
                        length_before = length_before + len(match[0])

        if re.search(pat_list['error_pattern'], line):
            errors = re.finditer(pat_list['error_pattern'], line) # list of match object of error pattern in line
            for error in errors:
                if ann.start() > error.start():
                    error_part = error[0].split(',')[0][1:]
                    length_before = length_before + (len(error[0]) - len(error_part))
            
        if re.search(pat_list['abs_er_pattern'], line):
            abs_ers = re.finditer(pat_list['abs_er_pattern'], line) # list of match object of abs_er pattern in line
            for abs_er in abs_ers:
                if ann.start() > abs_er.start():
                    length_before = length_before + 2

        if re.search(pat_list['note_pattern'], line):
            abs_errors = re.finditer(pat_list['note_pattern'], line) # list of match object of note pattern in line
            for abs_error in abs_errors:
                if ann.start() > abs_error.start():
                    length_before = length_before+1

        for pp in ['author_pattern', 'pecha_title_pattern', 'poti_title_pattern', 'chapter_title_pattern']:
            title_pattern = re.search(pat_list[pp], line)
            if title_pattern:
                if ann.start() > title_pattern.start():
                    length_before += 4
        
        for pp in ['start_cit_pattern', 'end_cit_pattern']:
            cit_patterns = re.finditer(pat_list[pp], line) # list of match object of citation pattern in line
            for cit_pattern in cit_patterns:
                if ann.start() > cit_pattern.start():
                    length_before = length_before+2

        for pp in ['start_sabche_pattern', 'end_sabche_pattern']:
            sabche_patterns = re.finditer(pat_list[pp], line) # list of match object of sabche pattern in line
            for sabche_pattern in sabche_patterns:
                if ann.start() > sabche_pattern.start():
                    length_before = length_before+2
        
        for pp in ['start_tsawa_pattern', 'end_tsawa_pattern']:
            tsawa_patterns = re.finditer(pat_list[pp], line) # list of match object of tsawa pattern in line
            for tsawa_pattern in tsawa_patterns:
                if ann.start() > tsawa_pattern.start():
                    length_before = length_before+2
        
        for pp in ['start_yigchung_pattern', 'end_yigchung_pattern']:
            yigchung_patterns = re.finditer(pat_list[pp], line) # list of match object of yigchung pattern in line
            for yigchung_pattern in yigchung_patterns:
                if ann.start() > yigchung_pattern.start():
                    length_before = length_before+2
        
        return length_before


    def base_extract(self, pat_list, annotated_line): 
        """ It extract the base text from annotated text.

        Args:
            pat_list (dict): It contains all the annotation's regex pattern as value and name of annotation as key.
            annotated_line (str): It contains the annotated line from which we want to extract the base text.
        
        Return:
            base_line (str): It contains the base text which is being extracted from the given annotated line.
        """
        base_line = annotated_line # stores the base_line which is line without annotation
        for pattern in ['line_pattern', 'topic_pattern', 'sub_topic_pattern', 'note_pattern']:
            base_line = re.sub(pat_list[pattern], '', base_line)
        
        for pattern in ['author_pattern', 'pecha_title_pattern', 'poti_title_pattern', 'chapter_title_pattern']:
            title_pattern = re.search(pat_list[pattern], annotated_line)
            if title_pattern:
                title = title_pattern[0][3:-1]
                base_line = re.sub(pat_list[pattern], title, base_line, 1)
        
        for pattern in ['start_cit_pattern', 'end_cit_pattern', 'start_sabche_pattern', 'end_sabche_pattern',
        'start_tsawa_pattern', 'end_tsawa_pattern', 'start_yigchung_pattern', 'end_yigchung_pattern']:
            base_line = re.sub(pat_list[pattern],'', base_line)

        if re.search(pat_list['error_pattern'], annotated_line):
            errors = re.finditer(pat_list['error_pattern'], annotated_line) # list of match object of error pattern in line
            for error in errors:
                error_part = error[0].split(',')[0][1:]
                base_line = re.sub(pat_list['error_pattern'], error_part, base_line, 1)
            
        if re.search(pat_list['abs_er_pattern'], annotated_line):
            abs_ers = re.finditer(pat_list['abs_er_pattern'], annotated_line)# list of match object of abs_er pattern in line
            for abs_er in abs_ers:
                base_line = re.sub(pat_list['abs_er_pattern'], abs_er[0][1:-1], base_line, 1)

        return base_line

    
    def build_layers(self, m_text, num_vol):
        
        i = 0  # tracker variable through out the text 

        cur_vol_pages = [] # list variable to store page annotation according to base string index eg : [(startPage,endPage)]
        cur_vol_error_id = [] # list variable to store error annotation according to base string index eg : [(es,ee,'suggestion')]
        cur_vol_abs_er_id = [] # list variable to store abs_er annotation 
        note_id = [] # list variable to store note annotation '#"
        pg_info = [] # lsit variable to store page info component
        pg_ann = [] # list variable to store page annotation content
        poti_titles = []
        chapter_titles = []
        start_cit_patterns = [] # list variable to store index of start citation pattern => (g
        end_cit_patterns = [] # list variable to store index of start citation pattern => g)
        start_sabche_pattern = [] # list variable to store index of start sabche pattern => (q
        end_sabche_pattern = [] # list variable to store index of end sabche pattern => q)
        start_tsawa_pattern = []  # list variable to store index of start tsawa pattern => (m
        end_tsawa_pattern = [] # list variable to store index of end sabche pattern => m)
        start_yigchung_pattern = [] # list variable to store index of start yigchung pattern => (y  
        end_yigchung_pattern = [] # list variable to store index of end yigchung pattern => y)
        
        pat_list={ 
            'author_pattern': r'\(au.+?\)',
            'pecha_title_pattern': r'\(k1.+?\)',
            'poti_title_pattern': r'\(k2.+?\)',
            'chapter_title_pattern': r'\(k3.+?\)',
            'page_pattern': r'\[[0-9]+[a-z]{1}\]',
            'line_pattern': r'\[\w+\.\d+\]','topic_pattern':r'\{\w+\}',
            'start_cit_pattern': r'\(g','end_cit_pattern': r'g\)',
            'start_sabche_pattern': r'\(q','end_sabche_pattern': r'q\)',
            'start_tsawa_pattern': r'\(m','end_tsawa_pattern': r'm\)',
            'start_yigchung_pattern': r'\(y','end_yigchung_pattern': r'y\)',
            'sub_topic_pattern': r'\{\w+\-\w+\}',
            'error_pattern': r'\(\S+\,\S+\)',
            'abs_er_pattern': r'\[[^0-9].*?\]',
            'note_pattern':r'#'
        }

        start_page = 0 # starting index of page
        end_page = 0 # ending index of page
        start_line = 0 #starting index of line
        end_line = 0 # ending index of line
        start_title = 0 # starting of the title component
        end_title = 0 # ending of the title component
        start_topic = 0 #starting index of topic_Id
        end_topic = 0 # ending index of topic_Id
        start_sub_topic = 0 #starting index of sub_topic_Id
        end_sub_topic = 0 #ending index of sub_topic_Id
        start_error = 0 #starting index of error
        end_error = 0 #ending index of error
        start_abs_er = 0 #starting index of abs_er
        end_abs_er = 0 #ending index of abs_er
        note = 0 #index of notes

        text_lines = m_text.splitlines() # list of all the lines in the text
        n_line = len(text_lines) # number of lines in the text 

        for idx, line in enumerate(text_lines):
                line = line.strip() 
                pat_len_before_ann = 0 # length of pattern recognised before  annotation
                if re.search(pat_list['page_pattern'], line):  # checking current line contains page annotation or not
                    start_page = end_page
                    end_page = end_line
                    page_info = line[re.search(pat_list['page_pattern'], line).end():]
                    pg_ann.append(re.search(pat_list['page_pattern'], line)[0][1:-1])
                    pg_info.append(page_info)
                    if len(pg_info)>=2:
                        cur_vol_pages.append((start_page, end_page, pg_info[-2], pg_ann[-2]))
                        if start_page < end_page: # to ignore the empty pages
                            i = i+1  # To accumulate the \n character 
                            end_page = end_page+3
                        self.base_text = self.base_text + '\n'
                elif re.search(pat_list['line_pattern'], line): #checking current line contains line annotation or not
                    start_line = i
                    length = len(line)

                    for pp in ['author_pattern','pecha_title_pattern','poti_title_pattern','chapter_title_pattern']:
                        title_pattern = re.search(pat_list[pp],line)
                        if title_pattern:
                            pat_len_before_ann = self.search_before(title_pattern, pat_list, line)
                            start_title = title_pattern.start()+i-pat_len_before_ann
                            end_title = start_title + len(title_pattern[0]) - 5
                            if pp == 'author_pattern':
                                self.author_pattern.append((start_title,end_title))
                            if pp == 'pecha_title_pattern':
                                self.pecha_title.append((start_title,end_title))
                            if pp == 'poti_title_pattern':
                                poti_titles.append((start_title,end_title))
                                end_topic = len(title_pattern[0][2:])
                                end_sub_topic = len(title_pattern[0][2:])
                            if pp == 'chapter_title_pattern':
                                chapter_titles.append((start_title,end_title))

                    if re.search(pat_list['sub_topic_pattern'], line): #checking current line contain sub_topicID annotation or not
                        sub_topic_match = re.search(pat_list['sub_topic_pattern'], line)
                        self.sub_topic_info.append(sub_topic_match[0][1:-1])
                        pat_len_before_ann = self.search_before(sub_topic_match, pat_list, line)
                        if start_sub_topic  == 0:
                            start_sub_topic = end_sub_topic
                            end_sub_topic = sub_topic_match.start()+i-pat_len_before_ann

                            if start_sub_topic < end_sub_topic:
                                if len(self.sub_topic_info) >= 2:
                                    self.sub_topic_Id.append((start_sub_topic, end_sub_topic, self.vol_walker+1, self.sub_topic_info[-2]))
                                    end_sub_topic = end_sub_topic
                                else:
                                    self.sub_topic_Id.append((start_sub_topic, end_sub_topic, self.vol_walker+1, self.sub_topic_info[-1]))
                                    end_sub_topic = end_sub_topic
                        else:
                            start_sub_topic = end_sub_topic
                            end_sub_topic = sub_topic_match.start()+i-pat_len_before_ann

                            if start_sub_topic < end_sub_topic:
                                self.sub_topic_Id.append((start_sub_topic, end_sub_topic, self.vol_walker+1, self.sub_topic_info[-2]))
                                end_sub_topic = end_sub_topic

                    if re.search(pat_list['topic_pattern'], line): #checking current line contain topicID annotation or not
                        topic = re.search(pat_list['topic_pattern'], line)
                        pat_len_before_ann = self.search_before(topic, pat_list, line)
                        self.topic_info.append(topic[0][1:-1])
                        start_topic = end_topic
                        end_topic = topic.start()+i-pat_len_before_ann

                        if (start_topic != end_topic or len(self.topic_info)>=2):
                            if len(self.topic_info) >= 2: # as we are ignoring the self.topic[0]
                                if start_topic < end_topic:
                                    self.current_topic_id.append((start_topic, end_topic, self.vol_walker+1, self.topic_info[-2])) # -2 as we need the secondlast item
                            else:
                                self.current_topic_id.append((start_topic, end_topic, self.vol_walker+1, self.topic_info[-1]))
                            self.topic_id.append(self.current_topic_id)
                            self.current_topic_id = []
                            if self.sub_topic_Id and end_sub_topic < end_topic:
                                self.sub_topic_Id.append((end_sub_topic,end_topic,self.vol_walker+1,self.sub_topic_info[-1]))
                            self.sub_topic.append(self.sub_topic_Id)
                            self.sub_topic_Id = []
                            if self.sub_topic_Id:
                                last = self.sub_topic_info[-1]
                                self.sub_topic_info = []
                                self.sub_topic_info.append(last)

                        
                    if re.search(pat_list['error_pattern'], line):   # checking current line contain error annotation or not
                        errors = re.finditer(pat_list['error_pattern'], line)
                        for error in errors:
                            suggestion = error[0].split(',')[1][:-1] # extracting the suggestion component
                            error_part = error[0].split(',')[0][1:]       # extracting the error component
                            pat_len_before_ann = self.search_before(error, pat_list, line)
                            start_error = error.start()+i-pat_len_before_ann

                            end_error = start_error+len(error_part)-1
                            cur_vol_error_id.append((start_error, end_error, suggestion))
                            
                    if re.search(pat_list['abs_er_pattern'], line):
                        abs_ers = re.finditer(pat_list['abs_er_pattern'],line)
                        for abs_er in abs_ers:
                            pat_len_before_ann=self.search_before(abs_er, pat_list, line)
                            start_abs_er = abs_er.start()+i-pat_len_before_ann
                            end_abs_er = start_abs_er + len(abs_er[0])-3
                            cur_vol_abs_er_id.append((start_abs_er,end_abs_er))
        
                    if re.search(pat_list['note_pattern'], line):
                        notes_in_line=re.finditer(pat_list['note_pattern'],line)
                        for notes in notes_in_line:
                            pat_len_before_ann=self.search_before(notes, pat_list, line)
                            note = notes.start()+i-pat_len_before_ann
                            note_id.append(note)
                    
                    if re.search(pat_list['start_cit_pattern'],line):
                        start_cits = re.finditer(pat_list['start_cit_pattern'],line)
                        for start_cit in start_cits:
                            pat_len_before_ann=self.search_before(start_cit,pat_list,line)
                            cit_start= start_cit.start()+i-pat_len_before_ann
                            start_cit_patterns.append(cit_start)

                    if re.search(pat_list['end_cit_pattern'],line):
                        end_cits = re.finditer(pat_list['end_cit_pattern'],line)
                        for end_cit in end_cits:
                            pat_len_before_ann=self.search_before(end_cit,pat_list,line)
                            cit_end= end_cit.start()+i-pat_len_before_ann-1
                            end_cit_patterns.append(cit_end)
                        
                    if re.search(pat_list['start_sabche_pattern'],line):
                        start_sabches = re.finditer(pat_list['start_sabche_pattern'],line)
                        for start_sabche in start_sabches:
                            pat_len_before_ann = self.search_before(start_sabche,pat_list,line)
                            sabche_start = start_sabche.start()+i-pat_len_before_ann
                            start_sabche_pattern.append(sabche_start)
                    
                    if re.search(pat_list['end_sabche_pattern'],line):
                        end_sabches = re.finditer(pat_list['end_sabche_pattern'],line)
                        for end_sabche in end_sabches:
                            pat_len_before_ann = self.search_before(end_sabche,pat_list,line)
                            sabche_end = end_sabche.start()+i-pat_len_before_ann-1
                            end_sabche_pattern.append(sabche_end)
                    
                    if re.search(pat_list['start_tsawa_pattern'],line):
                        start_tsawas = re.finditer(pat_list['start_tsawa_pattern'],line)
                        for start_tsawa in start_tsawas:
                            pat_len_before_ann = self.search_before(start_tsawa,pat_list,line)
                            tsawa_start = start_tsawa.start()+i-pat_len_before_ann
                            start_tsawa_pattern.append(tsawa_start)
                    
                    if re.search(pat_list['end_tsawa_pattern'], line):
                        end_tsawas = re.finditer(pat_list['end_tsawa_pattern'], line)
                        for end_tsawa in end_tsawas:
                            pat_len_before_ann = self.search_before(end_tsawa, pat_list, line)
                            tsawa_end = end_tsawa.start()+i-pat_len_before_ann-1
                            end_tsawa_pattern.append(tsawa_end)
                    
                    if re.search(pat_list['start_yigchung_pattern'], line):
                        start_yigchungs = re.finditer(pat_list['start_yigchung_pattern'], line)
                        for start_yigchung in start_yigchungs:
                            pat_len_before_ann = self.search_before(start_yigchung, pat_list, line)
                            yigchung_start = start_yigchung.start()+i-pat_len_before_ann
                            start_yigchung_pattern.append(yigchung_start)
                    
                    if re.search(pat_list['end_yigchung_pattern'], line):
                        end_yigchungs = re.finditer(pat_list['end_yigchung_pattern'], line)
                        for end_yigchung in end_yigchungs:
                            pat_len_before_ann = self.search_before(end_yigchung, pat_list, line)
                            yigchung_end = end_yigchung.start()+i-pat_len_before_ann-1
                            end_yigchung_pattern.append(yigchung_end)

                    pat_len_before_ann = self.total_pattern(pat_list, line)
                    end_line = start_line+length-pat_len_before_ann-1
                    i = end_line + 2
                    base_line = self.base_extract(pat_list, line) + '\n'
                    self.base_text += base_line
                    
                    if idx   ==  n_line-1:  # Last line case
                        start_page = end_page
                        start_topic = end_topic
                        start_sub_topic = end_sub_topic
                        if self.sub_topic_Id:
                            self.sub_topic_Id.append((start_sub_topic, i-2, self.vol_walker+1, self.sub_topic_info[-1] if self.sub_topic_info else None))
                        if self.topic_info:
                            self.current_topic_id.append((start_topic, i-2, self.vol_walker+1, self.topic_info[-1]))
                        cur_vol_pages.append((start_page, i-2, pg_info[-1], pg_ann[-1]))
                        self.page.append(cur_vol_pages)
                        self.error_id.append(cur_vol_error_id)
                        cur_vol_error_id = []
                        self.abs_er_id.append(cur_vol_abs_er_id)
                        cur_vol_abs_er_id = []
                        self.notes_id.append(note_id)
                        note_id = []
                        self.poti_title.append(poti_titles)
                        self.chapter_title.append(chapter_titles)
                        self.vol_walker += 1
        
        if num_vol == self.vol_walker: # checks the last volume
            self.topic_id.append(self.current_topic_id)
            self.current_topic_id = []
            self.sub_topic.append(self.sub_topic_Id)

        self.citation_pattern.append(self.merge(start_cit_patterns, end_cit_patterns)) # The starting and ending of  citation pattern is merged
        self.sabche_pattern.append(self.merge(start_sabche_pattern, end_sabche_pattern)) # The starting and ending of  sabche pattern is merged
        self.tsawa_pattern.append(self.merge(start_tsawa_pattern, end_tsawa_pattern)) # The starting and ending of  tsawa pattern is merged
        self.yigchung_pattern.append(self.merge(start_yigchung_pattern, end_yigchung_pattern)) # The starting and ending of  yigchung pattern is merged
    
    def get_result(self):

        if self.topic_id[0]:
            if self.topic_id[0][0][3] == self.topic_id[1][0][3]:
                self.topic_id = self.topic_id[1:]
                self.sub_topic = self.sub_topic[1:]
        self.sub_topic = self.__final_sub_topic(self.sub_topic)
        result = {
            'poti_title': self.poti_title,
            'chapter_title': self.chapter_title,
            'citation': self.citation_pattern,
            'page': self.page, # page variable format (start_index,end_index,pg_Info,pg_ann)
            'topic': self.topic_id,
            'sub_topic': self.sub_topic,
            'sabche': self.sabche_pattern,
            'tsawa': self.tsawa_pattern,
            'yigchung': self.yigchung_pattern,
            'correction': self.error_id,
            'error_candidate': self.abs_er_id,
            'peydurma': self.notes_id}
       
        return result

    def __final_sub_topic(self, sub_topics):
        """ It include all the sub topic belonging in one topic in a list.

        Args:
            sub_topic (list): It contains all the sub topic annotation's starting and ending index along with sub-topic info.

        Return:
            result (list): It contains list in which sub topic belonging to same topics are grouped together.
        """
        result = []
        cur_topic = []
        cur_sub = []
        sub_topic = sub_topics
        walker = 0;
        for i in range(0, len(sub_topic)):
            if len(sub_topic[i]) != 0:
                cur_sub.append(sub_topic[i][0])
                for walker in range(1, len(sub_topic[i])):
                    if sub_topic[i][walker][3] == sub_topic[i][walker-1][3]:
                        cur_sub.append(sub_topic[i][walker])
                    else:
                        cur_topic.append(cur_sub)
                        cur_sub = []
                        cur_sub.append(sub_topic[i][walker])
                if cur_sub:
                    cur_topic.append(cur_sub)
                    cur_sub = []       
            else:
                cur_topic.append(cur_sub)
                cur_sub = []
            result.append(cur_topic)
            cur_topic = []
        return result

    def get_base_text(self):
        base_text = self.base_text.strip()
        self.base_text = ''

        return base_text


    def create_opf(self,  input_path, **kwargs):
        input_path = Path(input_path)
        self._build_dirs(input_path)
        (self.dirs['opf_path']/'base').mkdir(exist_ok=True)

        for i, (m_text, vol_name, n_vol) in enumerate(self.get_input(input_path)):
            print(f'[INFO] Processing Vol {i+1:03} of {n_vol}: {vol_name} ...')
            base_id = f'v{i+1:03}'
            self.build_layers(m_text, n_vol)
            # save base_text
            # if (self.dirs['opf_path']/'base'/f'{base_id}.txt').is_file(): continue
            if 'is_text' in kwargs:
                if kwargs['is_text']:
                    continue

            base_text = self.get_base_text()
            (self.dirs['opf_path']/'base'/f'{base_id}.txt').write_text(base_text)

        # save pecha layers
        layers = self.get_result()
        for vol_layers, base_id in self.format_layer(layers):
            if base_id:
                print(f'[INFO] Creating layers for {base_id} ...')
                vol_layer_path = self.dirs['layers_path']/base_id
                vol_layer_path.mkdir(exist_ok=True)
            else:
                print('[INFO] Creating index layer for Pecha ...')

            for layer, ann in vol_layers.items():
                if layer == 'index':
                    layer_fn = self.dirs['opf_path']/f'{layer}.yml'
                else:
                    layer_fn = vol_layer_path/f'{layer}.yml'
                self.dump(ann, layer_fn)


class HFMLTextFromatter(HFMLFormatter):

    def __init_(self, output_path='./output', is_book=False):
        super().__init__(output_path=output_path, is_book=is_book)
        self.text_id = None

    def get_input(self, input_path):
        mtext = input_path.read_text()
        if self.is_book:
            return (self.text_preprocess(mtext), 'Book', 1)

        vol_pattern = r'\[v\d{3}\]'
        cur_vol_text =""
        vol_text = []
        vol_info = []
        vol_walker = 0
        lines = mtext.splitlines()
        n_line = len(lines)
        self.text_id = re.search(r'\{\w+\}',mtext)[0]
        for idx, line in enumerate(lines):
            vol_pat = re.search(vol_pattern,line)
            if vol_pat:
                vol_info.append(vol_pat[0][1:-1])
                if vol_walker > 0:
                    vol_text.append(cur_vol_text)
                    cur_vol_text = ""
                vol_walker += 1
            else:
                cur_vol_text += line +'\n'
            if idx == n_line-1:
                vol_text.append(cur_vol_text)
                cur_vol_text = ""
        
        result= []
        for i, vol_id in enumerate(vol_info):
            result.append((self.text_preprocess(vol_text[i]), vol_id, vol_walker))
        return result

    def __adapt_span_to_vol(self, extra, vol_walker):
        """ It adapts the index of parse output of serilized hfml file of a particular text id.

        Agrs:
            extra (int): start of text Id span which needs to be added to all the index of annotation present in that text id.
            vol_walker (int): Adapts all the annotation which contains volume information.
        """
        
        first_vol = []
        if self.poti_title[0]:
            start = self.poti_title[0][0] + extra
            end = self.poti_title[0][1] + extra
            first_vol.append((start, end))
            first_vol.append(self.poti_title[1:])
            self.poti_title = first_vol
            first_vol = []
        if self.chapter_title[0]:
            start = self.chapter_title[0][0] + extra
            end = self.chapter_title[0][1] + extra
            first_vol.append((start, end))
            first_vol.append(self.chapter_title[1:])
            self.poti_title = first_vol
            first_vol = []
        if self.citation_pattern[0]:
            for cit in self.citation_pattern[0]:
                start = cit[0]+extra
                end = cit[1]+extra
                first_vol.append((start, end))
            self.citation_pattern[0] =first_vol
            first_vol = []
        if self.sabche_pattern[0]:
            for sabche in self.sabche_pattern[0]:
                start = sabche[0]+extra
                end = sabche[1]+extra
                first_vol.append((start, end))
            self.sabche_pattern[0] =first_vol
            first_vol = []
        if self.yigchung_pattern[0]:
            for yig in self.yigchung_pattern[0]:
                start = yig[0]+extra
                end = yig[1]+extra
                first_vol.append((start, end))
            self.yigchung_pattern[0] =first_vol
            first_vol = []
        if self.tsawa_pattern[0]:
            for tsawa in self.tsawa_pattern[0]:
                start = tsawa[0]+extra
                end = tsawa[1]+extra
                first_vol.append((start, end))
            self.tsawa_pattern[0] =first_vol
            first_vol = []
        if self.error_id[0]:
            for cor in self.error_id[0]:
                start = cor[0]+extra
                end = cor[1]+extra
                first_vol.append((start, end, cor[2]))
            self.error_id[0] =first_vol
            first_vol = []
        if self.notes_id[0]:
            for cor in self.notes_id[0]:
                start = cor[0]+extra
                first_vol.append(start)
            self.notes_id[0] =first_vol
            first_vol = []
        if self.abs_er_id[0]:
            for er in self.abs_er_id[0]:
                start = er[0]+extra
                end = er[1]+extra
                first_vol.append((start, end))
            self.abs_er_id[0] =first_vol
            first_vol = []
        if self.page[0]:
            for pg in self.page[0]:
                start = pg[0]+extra
                end = pg[1]+extra
                first_vol.append((start, end, pg[2], pg[3]))
            self.page[0] =first_vol
            first_vol = []

        if self.topic_id:
            cur_top = []
            for cur_vol_top in self.topic_id[0]:
                start = cur_vol_top[0]
                end = cur_vol_top[1]
                if cur_vol_top[2] == 1:
                    start += extra
                    end += extra
                cur_top.append((start, end, cur_vol_top[2]+vol_walker, cur_vol_top[3]))
            self.topic_id[0] = cur_top

        if self.sub_topic[0][0]:
            cur_topic = []
            for st in self.sub_topic[0][0]:
                cur_sub_top = []
                for cst in st:
                    start = cst[0]
                    end = cst[1]
                    if cst[2] == 1:
                        start += extra
                        end += extra
                    cur_sub_top.append((start, end, t[2]+vol_walker, t[3]))
                cur_topic.append(cur_sub_top)
            self.sub_topic[0] = cur_topic

    def get_result(self):
        index = self.load(self.dirs['opf_path']/'index.yml')
        extra = 0
        for i, ann in enumerate(index['annotations']):
            if ann['work'] == self.text_id:
                extra = ann['work']['span'][0]['vol']['span']['start']
                break
        if not self.is_book:
            self.__adapt_span_to_vol(extra, i)

        result = {
            'poti_title': self.poti_title,
            'chapter_title': self.chapter_title,
            'citation': self.citation_pattern,
            'page': self.page, # page variable format (start_index,end_index,pg_Info,pg_ann)
            'topic': self.topic_id,
            'sub_topic': self.sub_topic,
            'sabche': self.sabche_pattern,
            'tsawa': self.tsawa_pattern,
            'yigchung': self.yigchung_pattern,
            'correction': self.error_id,
            'error_candidate': self.abs_er_id,
            'peydurma': self.notes_id}
       
        return result
if __name__ == "__main__":
    formatter = HFMLFormatter()
    formatter.create_opf('./tests/data/formatter/hfml/P000002/')

    formatter = HFMLTextFromatter()
    formatter.new_poti('./tests/data/formatter/hfml/vol_sep_test')



from copy import deepcopy
from pathlib import Path
from functools import partial

from bs4 import BeautifulSoup

from openpecha.formatters import BaseFormatter
from openpecha.formatters.format import *


class TsadraFormatter(BaseFormatter):
    '''
    OpenPecha Formatter for Tsadra DarmaCloud ebooks
    '''

    def __init__(self, output_path='./opfs'):
        super().__init__(output_path=output_path)
        self.base_text = ''
        self.walker = 0 # The walker to traverse every character in the pecha
        self.book_title = [] # list variable to store book title index
        self.author = [] # list variable to store author annotion index
        self.chapter = [] # list variable to store chapter annotation index
        self.root_text = [] # list variable to store root text index
        self.citation = [] # list variable to store citation index
        self.sabche = [] # list variable to store sabche index
        self.yigchung = [] # list variable to store yigchung index


    def text_preprocess(self, text):
        return text


    def format_layer(self, layers):
        for layer in layers:
            if layer == 'book_title':
                Book_title_layer = deepcopy(Layer)
                Book_title_layer['id'] = self.get_unique_id()
                Book_title_layer['annotation_type'] = 'book_title'
                Book_title_layer['revision'] = f'{1:05}'
                Book_title['span']['start'] = layers[layer][0][0]
                Book_title['span']['end'] = layers[layer][0][1]
                Book_title_layer['annotations'].append(Book_title)
            elif layer == 'author':
                Author_layer = deepcopy(Layer)
                Author_layer['id'] = self.get_unique_id()
                Author_layer['annotation_type'] = 'author'
                Author_layer['revision'] = f'{1:05}'
                for a in layers[layer]:
                    author = deepcopy(Author)
                    author['span']['start'] = a[0]
                    author['span']['end'] = a[1]
                    Author_layer['annotations'].append(author)
            elif layer == 'chapter_title':
                Chapter_layer = deepcopy(Layer)
                Chapter_layer['id'] = self.get_unique_id()
                Chapter_layer['annotation_type'] = 'chapter_title'
                Chapter_layer['revision'] = f'{1:05}'
                for a in layers[layer]:
                    chapter = deepcopy(Chapter)
                    chapter['id'] = self.get_unique_id()
                    chapter['span']['start'] = a[0]
                    chapter['span']['end'] = a[1]
                    Chapter_layer['annotations'].append(chapter)
            elif layer == 'tsawa':
                Tsawa_layer = deepcopy(Layer)
                Tsawa_layer['id'] = self.get_unique_id()
                Tsawa_layer['annotation_type'] = 'tsawa'
                Tsawa_layer['revision'] = f'{1:05}'
                for a in layers[layer]:
                    tsawa = deepcopy(Tsawa)
                    tsawa['id'] = self.get_unique_id()
                    tsawa['span']['start'] = a[0]
                    tsawa['span']['end'] = a[1]
                    Tsawa_layer['annotations'].append(tsawa)
            elif layer == 'quotation':
                Quotation_layer = deepcopy(Layer)
                Quotation_layer['id'] = self.get_unique_id()
                Quotation_layer['annotation_type'] = 'tsawa'
                Quotation_layer['revision'] = f'{1:05}'
                for a in layers[layer]:
                    quote = deepcopy(Quotation)
                    quote['id'] = self.get_unique_id()
                    quote['span']['start'] = a[0]
                    quote['span']['end'] = a[1]
                    Quotation_layer['annotations'].append(quote)
            elif layer == 'sabche':
                Sabche_layer = deepcopy(Layer)
                Sabche_layer['id'] = self.get_unique_id()
                Sabche_layer['annotation_type'] = 'tsawa'
                Sabche_layer['revision'] = f'{1:05}'
                for a in layers[layer]:
                    sabche = deepcopy(Sabche)
                    sabche['id'] = self.get_unique_id()
                    sabche['span']['start'] = a[0]
                    sabche['span']['end'] = a[1]
                    Sabche_layer['annotations'].append(sabche)
            elif layer == 'yigchung':
                Yigchung_layer = deepcopy(Layer)
                Yigchung_layer['id'] = self.get_unique_id()
                Yigchung_layer['annotation_type'] = 'tsawa'
                Yigchung_layer['revision'] = f'{1:05}'
                for a in layers[layer]:
                    yigchung = deepcopy(Yigchung)
                    yigchung['id'] = self.get_unique_id()
                    yigchung['span']['start'] = a[0]
                    yigchung['span']['end'] = a[1]
                    Yigchung_layer['annotations'].append(yigchung)
        
        result = {
            'book_title': Book_title_layer,
            'author': Author_layer,
            'chapter_title': Chapter_layer,
            'tsawa': Tsawa_layer,
            'quotation': Quotation_layer,
            'sabche': Sabche_layer,
            'yigchung': Yigchung_layer
        }

        return result


    def build_layers(self, html):
        '''
        To Build the layer
        '''
        soup = BeautifulSoup(html, 'html.parser')
        book_title_tmp = ''
        author_tmp = ''
        chapter_title_tmp = ''
        root_text_tmp = ''
        sabche_tmp = ''
        commentary_tmp = ''
        citation_tmp = ''
        rt_base = 'tibetan-root-text_tibetan-root-text'
        cit_base = 'tibetan-citations-in-verse_'
        com_base = 'tibetan-commentary'
        com_classes = {
            'first': f'{com_base}-first-line',
            'middle': f'{com_base}-middle-lines',
            'last': f'{com_base}-last-line',
            'non': f'{com_base}-non-indent1'
        }
        cit_classes = {
            'first': f'{cit_base}tibetan-citations-first-line',
            'middle': f'{cit_base}tibetan-citations-middle-lines',
            'last': f'{cit_base}tibetan-citations-last-line',
            'indent': f'{cit_base}tibetan-regular-indented'
        }
        root_text_classes = {
            'first': f'{rt_base}-first-line',
            'middle': f'{rt_base}-middle-lines',
            'last': f'{rt_base}-last-line'
        }
        #p_with_citations = []
        ps = soup.find_all('p')
        for p in ps:
            if 'credits-page_front-title' in  p['class'][0]: # to get the book title index
                book_title_tmp = self.text_preprocess(p.text) + '\n'
                self.book_title.append((self.walker, len(book_title_tmp)-2+self.walker))
                self.base_text += book_title_tmp
                self.walker += len(book_title_tmp)

            if 'text-author' in p['class'][0]: # to get the author annotation index
                author_tmp = self.text_preprocess(p.text) + '\n'
                self.author.append((self.walker, len(author_tmp)-2+self.walker))
                self.base_text += author_tmp
                self.walker += len(author_tmp)

            if rt_base in p['class'][0]: # to get the root text or 'tsawa' annotation index (verse form)
                #TODO: one line root text.   
                if p['class'][0] == root_text_classes['first'] or \
                p['class'][0] == root_text_classes['middle']:
                    root_text_tmp += self.text_preprocess(p.text) + '\n'
                    self.base_text += self.text_preprocess(p.text) + '\n'
                elif p['class'][0] == root_text_classes['last']:
                    for s in p.contents:
                        if 'root' in s['class'][0]:
                            root_text_tmp += self.text_preprocess(s.text)
                            self.root_text.append((self.walker,len(root_text_tmp)-1+self.walker))
                            self.walker+= len(root_text_tmp) + 1
                            root_text_tmp = ''
                        else:
                            self.walker+= len(self.text_preprocess(s.text))
                    self.base_text += self.text_preprocess(p.text) + '\n'

            elif 'tibetan-chapter' in p['class'][0]: # to get chapter title index
                chapter_title_tmp = self.text_preprocess(p.text) +'\n'
                self.chapter.append((self.walker, len(chapter_title_tmp)-2+self.walker))
                self.walker+= len(chapter_title_tmp)
                self.base_text += chapter_title_tmp

            elif 'commentary' in p['class'][0] or 'tibetan-regular-indented' in p['class'][0]:

                # travesing through commetary which are in verse form
                if p['class'][0] == com_classes['first'] or \
                p['class'][0] == com_classes['middle']:
                    commentary_tmp += self.text_preprocess(p.text) + '\n'
                    self.base_text += self.text_preprocess(p.text) + '\n'
                elif p['class'][0] == com_classes['last']:
                    commentary_tmp += self.text_preprocess(p.text) + '\n'
                    self.base_text += self.text_preprocess(p.text) + '\n'
                    self.walker+= len(commentary_tmp)
                    commentary_tmp = ''
                #travesing through each span of commentary and regular ptag to search annotations
                else:            
                    p_tmp = ''
                    p_walker=self.walker
                    for s in p.contents:
                        #check for page with no content and skip the page
                        if isinstance(s, str): return '' 
                        #some child are not <span> rather like <a> and some <span> has no 'class' attr
                        try:
                            s['class'][0]
                        except:
                            p_tmp += self.text_preprocess(s.text)
                            continue

                        if 'small' in s['class'][0]: # checking for yigchung annotation
                        # p_tmp += f'{{++*++}}{self.text_preprocess(s.text)}{{++*++}}'
                            if citation_tmp:
                            #citation_tmp += citation_tmp
                                self.citation.append((p_walker,len(citation_tmp)-2+p_walker))
                                p_walker += len(citation_tmp)
                                citation_tmp = ''
                            self.yigchung.append((p_walker,len(self.text_preprocess(s.text))-1+p_walker))
                            p_walker += len(self.text_preprocess(s.text))

                        elif 'external-citations' in s['class'][0]: # checking for citation annotation
                            citation_tmp += self.text_preprocess(s.text)
                        
                        elif 'front-title' in s['class'][0]:
                            if citation_tmp:
                                self.citation.append((p_walker,len(citation_tmp)-2+p_walker))
                                p_walker += len(citation_tmp)
                                citation_tmp = ''
                            p_walker += len(self.text_preprocess(s.text))
                        else:
                            if citation_tmp:
                                self.citation.append((p_walker,len(citation_tmp)-2+p_walker))
                                p_walker += len(citation_tmp)
                                citation_tmp = ''
                            p_walker += len(self.text_preprocess(s.text))

                    #when citation ends the para
                    if citation_tmp: 
                        self.citation.append((p_walker,len(citation_tmp)-2+p_walker))
                        p_walker += len(citation_tmp)
                        citation_tmp = ''
                

                    commentary_tmp = self.text_preprocess(p.text) + '\n'
                    self.base_text += self.text_preprocess(commentary_tmp)
                    self.walker+= len(commentary_tmp)
                    commentary_tmp = ''
                    p_walker = 0

            elif 'sabche' in p['class'][0]: # checking for sabche annotation
                sabche_tmp = ''
                p_with_sabche_tmp = ''
                k = self.walker
                for s in p.contents:
                    try:
                        s['class'][0]
                    except:
                        p_with_sabche_tmp += self.text_preprocess(s.text)
                        continue

                    if 'sabche' in s['class'][0]:
                        sabche_tmp += self.text_preprocess(s.text)
                    
                    elif 'front-tile' in s['class'][0]:
                        k += len(self.text_preprocess(s.text))
                

                #when sabche ends the para
                if sabche_tmp:
                        self.sabche.append((k,len(sabche_tmp)-1+k))
                        sabche_tmp=''             
                self.walker+= len(self.text_preprocess(p.text))+1
                self.base_text += self.text_preprocess(p.text) + '\n'
                k = 0
            elif cit_base in p['class'][0]: # checking for citation annotation first two if for verse form and last for non verse
                if p['class'][0] == cit_classes['first'] or \
                    p['class'][0] == cit_classes['middle']:
                    citation_tmp += self.text_preprocess(p.text) + '\n'
                    self.base_text += self.text_preprocess(p.text)  + '\n'
                elif p['class'][0] == cit_classes['last']:
                    citation_tmp += self.text_preprocess(p.text) + '\n'
                    self.citation.append((self.walker,len(citation_tmp)-2+self.walker))
                    self.base_text += self.text_preprocess(p.text) + '\n'
                    self.walker+= len(citation_tmp)
                    citation_tmp = ''
                elif p['class'][0] == cit_classes['indent']:
                    citation_tmp += self.text_preprocess(p.text) + '\n'
                    self.base_text += self.text_preprocess(p.text) + '\n'
                    self.citation.append(self.walker, len(citation_tmp)-2+self.walker)
                    self.walker+= len(citation_tmp)
                    citation_tmp = ''


    def get_result(self):
        '''
        To return all the result
        '''
        result = {
            'book_title': self.book_title,
            'author': self.author,
            'chapter_title': self.chapter,
            'tsawa': self.root_text,
            'quotation': self.citation,
            'sabche': self.sabche,
            'yigchung': self.yigchung
        }
        return result


    def _get_meta_data(ann):
            return ','.join([self.base_text[a[0]: a[1]+1] for a in ann])


    def get_base_text(self):
        '''
        To return base text of each processed page
        '''
        return self.base_text


    def get_input(self, input_path):
        
        def get_prefix(html_paths):
            self.sku = list(html_paths[0].parents)[1].name
            return sorted(html_paths, key=lambda p: len(p.stem))[0].stem

        def semantic_order(sku, html_path):
            html_fn = html_path.stem
            order = html_fn[len(sku)+1:]
            if order:
                return int(order)
            else:
                return 0

        html_paths = [o for o in input_path.iterdir() if o.suffix == '.xhtml' and o.stem != 'cover']
        sku_sementic_order = partial(semantic_order, get_prefix(html_paths))
        html_paths = sorted(html_paths, key=sku_sementic_order)
        html_paths.insert(0, input_path/'cover.xhtml')

        for html_fn in html_paths:
            yield Path(html_fn).read_text()


    def create_metadata(self, layers):

        def get_text(span):
            return self.base_text[span[0]: span[1]+1].replace('\n', '')

        meta_data = {}
        meta_data['title'] = get_text(layers['book_title'][0])
        meta_data['authors'] = [get_text(span) for span in layers['author']]
        meta_data['sku'] = self.sku
        meta_data['layers'] = [l for l in layers if layers[l]]
        return {'ebook_metadata': meta_data}


    def create_opf(self, input_path, id):
        input_path = Path(input_path)
        self._build_dirs(input_path, id=id)
        (self.dirs['opf_path']/'base').mkdir(exist_ok=True)

        # parse layers
        for html in self.get_input(input_path):
            self.build_layers(html)
        
        # save base-text
        (self.dirs['opf_path']/'base'/f'v001.txt').write_text(self.get_base_text())

        # format and save layer
        vol_layer_path = self.dirs['layers_path']/'v001'
        vol_layer_path.mkdir(exist_ok=True)
        layers = self.get_result()
        for layer, ann in self.format_layer(layers).items():
            if ann['annotations']: 
                layer_fn = vol_layer_path/f'{layer}.yml'
                self.dump(ann, layer_fn)

        # save metatdata
        meta_data = self.create_metadata(layers)
        meta_fn = self.dirs['opf_path']/'meta.yml'
        self.dump(meta_data, meta_fn)

if __name__ == "__main__":
    path = 'bo_crawler/bo_crawler/data/tsadra/data/ebooks/IBA-LG-04-1/OEBPS/'
    formatter = TsadraFormatter(output_path='./test_opf')
    formatter.create_opf(path, 1)



import os
from pathlib import Path
import time

from github import Github
from git import Repo


g = Github(os.environ.get('GITHUB_TOKEN'))
org = g.get_organization('OpenPecha')

def create_github_repo(path):
    repo = org.create_repo(path.name)
    time.sleep(2)
    return repo._html_url.value
    

def commit(repo, message, not_includes):
    has_changed = False

    # add untrack fns
    for fn in repo.untracked_files:

        ignored = False
        for not_include_fn in not_includes:
            if not_include_fn in fn:
                ignored = True
        
        if ignored:
            continue

        if fn: repo.git.add(fn)
        if has_changed is False:
            has_changed = True

    # add modified fns
    if repo.is_dirty() is True:
        for fn in repo.git.diff(None, name_only=True).split('\n'):
            if fn: repo.git.add(fn)
            if has_changed is False:
                has_changed = True
    
    if has_changed is True:
        if not message:
            message = 'Initial commit'
        repo.git.commit('-m', message)
        repo.git.push('origin', 'master')


def create_local_repo(path, remote_url):
    if (path/'.git').is_dir():
        return Repo(path)
    else:
        repo = Repo.init(path)
        old_url = remote_url.split('//')
        usr, pw = os.environ['GH_USR'], os.environ['GH_PW']
        auth_remote_url = f'{old_url[0]}//{usr}:{pw}@{old_url[1]}'
        repo.create_remote('origin', auth_remote_url)
        return repo


def github_publish(path, message=None, type=None, not_includes=None):
    path = Path(path)
    remote_repo_url = create_github_repo(path)
    local_repo = create_local_repo(path, remote_repo_url)
    commit(local_repo, message, not_includes)


def create_file(repo_name, path, content, msg, update=False):
    repo = org.get_repo(repo_name)
    if update:
        old_content = repo.get_contents(path)
        repo.update_file(old_content.path, msg, content, old_content.sha)
    else:
        repo.create_file(path, msg, content)

def get_bump_tag(repo):
    try:
        latest_release_tag = repo.get_latest_release().tag_name
    except:
        return 'v0.1'

    tag_number = round(float(latest_release_tag[1:]), 1)
    bump_tag_number = tag_number + 0.1
    return f'v{bump_tag_number}'


def create_release(repo_name, asset_paths=None):
    repo = org.get_repo(repo_name)
    bump_tag = get_bump_tag(repo)
    new_release = repo.create_git_release(
        bump_tag, bump_tag, 'Initial Release'
    )
    for asset_path in asset_paths:
        new_release.upload_asset(asset_path)


def create_readme(metadata, path):
    result = ''
    # add title
    result += f"## Title\n\t- {metadata['title']}\n\n"

    # add author
    result += f"## Author\n\t- {metadata['author']}\n\n"

    readme_fn = path/'README.md'
    readme_fn.write_text(result)


def delete_repo(repo_name):
    repo = org.get_repo(repo_name)
    repo.delete()


if __name__ == "__main__":
    create_release('P000002', asset_paths='./output/P000300/release.zip')


from openpecha.serializers import Serialize


class SerializeHFML(Serialize):
    """
    HFML (Human Friendly Markup Language) serializer class for OpenPecha.
    """

    def apply_annotation(self, vol_id, ann):
        only_start_ann = False
        start_payload = '('
        end_payload = ')'
        side = 'ab'
        if ann['type'] == 'pagination':
            if ann['page_index'] == '0b':
                pg_n = ann['reference'][5:-1]
                pg_side = ann['reference'][-1]
                if '-' in pg_n:
                    pg_n = int(pg_n.split('-')[0])
                    pg_side = side[int(pg_side)]
                    start_payload = f'[{pg_n}{pg_side}]'
                else:
                    pg_n = int(pg_n)
                    if pg_side.isdigit():
                        pg_n = str(pg_n) + pg_side
                        pg_side = ''
                    start_payload = f'[{pg_n}{pg_side}]'
            else:
                start_payload = f'[{ann["page_index"]}]'
            
            if ann["page_info"]:
                start_payload += f' {ann["page_info"]}\n'
            else:
                start_payload += f' {ann["reference"]}\n'
            only_start_ann = True
        elif ann['type'] == 'correction':
            start_payload = '('
            end_payload = f',{ann["correction"]})'
        elif ann['type'] == 'peydurma':
            start_payload = '#'
            only_start_ann = True
        elif ann['type'] == 'error_candidate':
            start_payload = '['
            end_payload = ']'
        elif ann['type'] == 'book_title':
            start_payload = '(k1'
            end_payload = ')'
        elif ann['type'] == 'author':
            start_payload = '(au'
            end_payload = ')'
        elif ann['type'] == 'chapter_title':
            start_payload = '(k3'
            end_payload = ')'
        elif ann['type'] == 'tsawa':
            start_payload = '(m'
            end_payload = 'm)'
        elif ann['type'] == 'quotation':
            start_payload = '(g'
            end_payload = 'g)'
        elif ann['type'] == 'sabche':
            start_payload = '(q'
            end_payload = 'q)'
        elif ann['type'] == 'yigchung':
            start_payload = '(y'
            end_payload = 'y)'
        
        start_cc, end_cc = self.__get_adapted_span(ann['span'], vol_id)
        #start_cc -= 4
        self.add_chars(vol_id, start_cc, True, start_payload)
        if not only_start_ann:
            self.add_chars(vol_id, end_cc, False, end_payload)


from copy import deepcopy
import gzip
import json
import math
from pathlib import Path
import re
import yaml

from openpecha.formatters import BaseFormatter
from openpecha.formatters.format import Layer
from openpecha.formatters.format import Page
from openpecha.utils import gzip_str


class GoogleOCRFormatter(BaseFormatter):
    '''
    OpenPecha Formatter for Google OCR JSON output of scanned pecha.
    '''

    def __init__(self, output_path='./output'):
        super().__init__(output_path=output_path)
        self.n_page_breaker_char = 3
        self.page_break = '\n' * self.n_page_breaker_char
        self.base_text = []


    def text_preprocess(self, text):
        
        return text

    
    def get_input(self, input_path):
        '''
        load and return all jsons in the input_path.
        '''
        for fn in sorted(list(input_path.iterdir())):
            if fn.name.split('.')[0] == 'info': continue
            try:
                yield json.load(gzip.open(str(fn), 'rb')), fn.name.split('.')[0]
            except GeneratorExit:
                return None, None
            except:
                yield None, None


    def _get_page_index(self, n):
        page_sides = 'ab'
        div = n/2
        page_num = math.ceil(div)
        if div.is_integer():
            return f'{page_num}{page_sides[1]}'
        else:
            return f'{page_num}{page_sides[0]}'


    def format_layer(self, layers, base_id):
        # Format page annotation
        Pagination = deepcopy(Layer)
        Pagination['id'] = self.get_unique_id()
        Pagination['annotation_type'] = 'pagination'
        Pagination['revision'] = f'{1:05}'
        for pg, page_ref in zip(layers['pages'], layers['pages_ref']):
            page = deepcopy(Page)
            page['id'] = self.get_unique_id()
            page['span']['start'] = pg[0]
            page['span']['end'] = pg[1]
            page['page_index'] = self._get_page_index(pg[2])
            page['reference'] = page_ref
            Pagination['annotations'].append(page)

        result = {
            'pagination': Pagination
        }

        return result


    def _get_coord(self, vertices):
        coord = []
        for vertice in vertices:
            coord.append((vertice['x'], vertice['y']))
        
        return coord


    def _get_page(self, response):
        try:
            page = response['textAnnotations'][0]
        except KeyError:
            return None, None
        
        text = page['description']
        # vertices = page['boundingPoly']['vertices']  # get text box
        
        return text, None #self._get_coord(vertices)


    def _get_lines(self, text, last_pg_end_idx, first_pg):
        lines = []
        line_breaks = [m.start() for m in re.finditer('\n', text)]
        
        start = last_pg_end_idx
        
        # increase the start idx with page_breaker_char for page greater than frist page.
        if not first_pg:
            start += self.n_page_breaker_char+1
            line_breaks = list(map(lambda x: x+start, line_breaks))

        for line in line_breaks:
            lines.append((start, line-1)) # skip new_line, which has 1 char length
            start += (line-start) + 1
        
        return lines, line


    def save_boundingPoly(self, response, path):

        def tlbr(vertices):
            return {'tl': vertices[0], 'br': vertices[2]}

        boundingPolies = {}
        for ann in response['textAnnotations']:
            boundingPolies[ann['description']] = tlbr(ann['boundingPoly']['vertices'])

        zipped_boundingPolies = gzip_str(json.dumps(boundingPolies))
        path.write_bytes(zipped_boundingPolies)

    def save_low_conf_char(self, response, path):
        char_idx = 0
        low_conf_chars = ''
        for page in response['fullTextAnnotation']['pages']:
            for block in page['blocks']:
                for paragraph in block['paragraphs']:
                    for word in paragraph['words']:
                        for symbol in word['symbols']:
                            if symbol.get('confidence', 1) < 0.9:
                                low_conf_chars += f'{symbol["text"]} {char_idx}\n'
                            char_idx += 1
        if low_conf_chars:
            path.write_bytes(gzip_str(low_conf_chars))


    def build_layers(self, responses, vol_name, base_id=None):
        if base_id:
            bounding_poly_vol_path = self.dirs['release_path']/'bounding_poly'/base_id
            low_conf_char_vol_path = self.dirs['release_path']/'low_confidence_chars'/base_id
            bounding_poly_vol_path.mkdir(parents=True, exist_ok=True)
            low_conf_char_vol_path.mkdir(parents=True, exist_ok=True)

        pages = []
        pages_ref = []
        last_pg_end_idx = 0
        for response, page_ref in responses:
            # extract page number, eg: I1PD901350083 -> 83
            n_pg_str = page_ref[len(vol_name):]
            if '-' in page_ref:
                n_pg_str = page_ref.split('-')[-1]

            if n_pg_str and n_pg_str[-1].isalpha():
                n_pg_str = n_pg_str[:-1]

            try:
                n_pg = int(n_pg_str)
            except:
                #TODO: fix later, collection all the cases as of now
                n_pg = 0 # dummy value

            # extract annotation
            if not response:
                print(f'[ERROR] Failed : {n_pg}')
                continue
            text, _ = self._get_page(response)

            # skip empty page (can be bad image)
            if not text:
                print(f'[ERROR] empty page {n_pg}')
                continue
            lines, last_pg_end_idx = self._get_lines(text, last_pg_end_idx, n_pg == 1)
            pages.append((lines[0][0], lines[-1][1], n_pg))
            pages_ref.append(response.get('image_link', page_ref))

            # create base_text
            self.base_text.append(text)

            if base_id:
                # save the boundingPoly to resources
                self.save_boundingPoly(response, bounding_poly_vol_path/f'{n_pg:04}.json.gz')
                # save low confident char and it's corresponding index
                self.save_low_conf_char(response, low_conf_char_vol_path/f'{n_pg:04}.txt.gz')

        result = {
            'pages': pages,
            'pages_ref': pages_ref,
        }
            
        return result


    def get_base_text(self):
        base_text = f'{self.page_break}'.join(self.base_text)
        self.base_text = []

        return base_text


    def get_metadata(self, work_id):
        import requests
        import xml.etree.ElementTree as ET
        from pyewts import pyewts
        
        converter = pyewts()
        query_url = 'https://www.tbrc.org/xmldoc?rid={}'
        bdrc_metadata_url = query_url.format(work_id)
        r = requests.get(bdrc_metadata_url)
        root = ET.fromstring(r.content.decode('utf-8'))        
        title_tag = root[0]
        author_tag = root.find('{http://www.tbrc.org/models/work#}creator')
        metadata = {
            'id': f'opecha:{self.pecha_id}',
            'initial_creation_type': 'ocr',
            'source_metadata': {
                'id': f'bdr:{work_id}',
                'title': converter.toUnicode(title_tag.text),
                'volume': '',
                'author': converter.toUnicode(author_tag.text) if author_tag else '',
            }
        }

        return metadata


    def create_opf(self, input_path, pecha_id):
        input_path = Path(input_path)
        self._build_dirs(input_path, id=pecha_id)

        self.dirs['release_path'] = self.dirs['opf_path'].parent/'releases'
        self.dirs['release_path'].mkdir(exist_ok=True, parents=True)

        for i, vol_path in enumerate(sorted(input_path.iterdir())):
            print(f'[INFO] Processing {input_path.name}-{vol_path.name} ...')
            base_id = f'v{i+1:03}'
            if (self.dirs['opf_path']/'base'/f'{base_id}.txt').is_file(): continue
            responses = self.get_input(vol_path)
            layers = self.build_layers(responses, vol_path.name, base_id)
            formatted_layers = self.format_layer(layers, base_id)
            base_text = self.get_base_text()

            # save base_text
            (self.dirs['opf_path']/'base'/f'{base_id}.txt').write_text(base_text)

            # save layers
            vol_layer_path = self.dirs['layers_path']/base_id
            vol_layer_path.mkdir(exist_ok=True)
            for layer, ann in formatted_layers.items():
                layer_fn = vol_layer_path/f'{layer}.yml'
                self.dump(ann, layer_fn)

        # create meta.yml
        meta_fn = self.dirs['opf_path']/'meta.yml'
        self.dump(self.get_metadata(input_path.name), meta_fn)

        return self.dirs['opf_path'].parent


if __name__ == "__main__":
    formatter = GoogleOCRFormatter()
    formatter.create_opf('./tests/data/formatter/google_ocr/W3CN472', 300)


'''Utilities functions
'''
import gzip
import io
import shutil

from openpecha.github_utils import create_release


def gzip_str(string_):
    # taken from https://gist.github.com/Garrett-R/dc6f08fc1eab63f94d2cbb89cb61c33d
    out = io.BytesIO()

    with gzip.GzipFile(fileobj=out, mode='w') as fo:
        fo.write(string_.encode())

    bytes_obj = out.getvalue()
    return bytes_obj


def ocr_result_input(path):
    return path


def create_release_with_assets(path):
    asset_paths = []
    for asset_path in (path/'releases').iterdir():
        shutil.make_archive(asset_path, 'zip', asset_path)
        asset_paths.append(f'{str(asset_path)}.zip')

    create_release(path.name, asset_paths=asset_paths)


import csv
import requests
from pathlib import Path
import shutil

import click
from github import Github
from git import Repo
from tqdm import tqdm

from openpecha.serializers import *
from openpecha.blupdate import Blupdate
from openpecha.formatters import *


OP_PATH = Path('./.openpecha')
config = {
    # Github
    'OP_CATALOG_URL': 'https://raw.githubusercontent.com/OpenPoti/openpecha-catalog/master/data/catalog.csv',
    'OP_ORG': 'https://github.com/OpenPecha',

    # Local
    'OP_DATA_PATH': (OP_PATH/'data').resolve(),
    'OP_CATALOG_PATH': (OP_PATH/'data'/'catalog.csv').resolve(),
    'CONFIG_PATH': (OP_PATH/'config').resolve(),
    'DATA_CONFIG_PATH': (OP_PATH/'config'/'data_config').resolve(),
}

ERROR = '[ERROR] {}'
INFO = '[INFO] {}'

def get_pecha_id(n):
    return f'P{int(n):06}'

@click.group()
def cli():
    pass


def create_config_dirs():
    config['OP_DATA_PATH'].mkdir(parents=True, exist_ok=True)
    config['CONFIG_PATH'].mkdir(parents=True, exist_ok=True)


def get_pecha(id, batch_path, layers):

    def _check_pecha(id=None, pechas=None, layer=None, pecha_list=None):
        if id not in pecha_list:
            if id in pechas:
                if layer:
                    if layer in pechas[id][-1].split('_'):
                        pecha_list.append(id)
                    else:
                        msg = f'{layer} layer is not found in {id}'
                        click.echo(ERROR.format(msg))
                else:
                    pecha_list.append(id)
            else:
                msg = f'{id} not found in OpenPecha catalog'
                click.echo(ERROR.format(msg))

    def _get_batch(batch_path):
        with Path(batch_path).open() as f:
            batch_ids = [line.strip() for line in f.readlines()]
        return batch_ids


    pecha_list = []

    # If filter by layers
    if layers:
        layers_name = [l.strip() for l in layers.split(',')]
        for layer in layers_name:
            batch_ids = None
            if id:
                _check_pecha(id=id, pechas=pechas, layer=layer, pecha_list=pecha_list)
            elif batch_path:
                if not batch_ids: batch_ids = _get_batch(batch_path)
                for b_id in batch_ids:
                    _check_pecha(id=b_id, pechas=pechas, layer=layer, pecha_list=pecha_list)
            else:
                for p_id in pechas:
                    _check_pecha(id=p_id, pechas=pechas, layer=layer, pecha_list=pecha_list)
    else:
        if id:
            _check_pecha(id=id, pechas=pechas, pecha_list=pecha_list)
        elif batch_path:
            batch_ids = _get_batch(batch_path)
            for b_id in batch_ids:
                _check_pecha(id=b_id, pechas=pechas, pecha_list=pecha_list)
        else:
            for p_id in pechas:
                _check_pecha(id=p_id, pechas=pechas, pecha_list=pecha_list)

    return pecha_list


def download_pecha(pecha, out):
    # clone the repo
    pecha_url = f"{config['OP_ORG']}/{pecha}.git"
    pecha_path = config['OP_DATA_PATH']/pecha
    if pecha_path.is_dir(): # if repo is already exits at local then try to pull
        repo = Repo(str(pecha_path))
        repo.heads['master'].checkout()
        repo.remotes.origin.pull()
    else:
        Repo.clone_from(pecha_url, str(pecha_path))


# Poti Download command
@cli.command()
@click.option('--number', '-n', help='Pecha number of pecha, for single pecha download')
@click.option('--batch', '-b', help="path to text file containg list of names of \
                                     pecha in separate line. Poti batch download")
@click.option('--filter', '-f', help='filter pecha by layer availability, specify \
                                     layer names in comma separated, eg: title,yigchung,..')
@click.option('--out', '-o', default='./pecha',
                            help='directory to store all the pecha')
def download(**kwargs):
    '''
    Command to download pecha.
    If id and batch options are not provided then it will download all the pecha.
    '''
    pecha_id = get_pecha_id(kwargs['number'])

    # create config dirs
    create_config_dirs()

    # configure the data_path
    config['data'] = Path(kwargs['out']).resolve()

    # get pecha
    # pechas = get_pecha(work_id, kwargs['batch'], kwargs['filter'])
    pechas = [pecha_id]

    # download the repo
    for pecha in tqdm(pechas):
        download_pecha(pecha, kwargs['out'])

    # save data_path in data_config
    config_path = config['DATA_CONFIG_PATH']
    if not config_path.is_file():
        config_path.write_text(str(config['data'].resolve()))

    # print location of data
    msg = f'Downloaded {pecha_id} ... ok'
    click.echo(INFO.format(msg))


# Apply layer command
layers_name = ['title', 'tsawa', 'yigchung', 'quotes', 'sapche']

@cli.command()
@click.option('--name', '-n', type=click.Choice(layers_name), \
                              help='name of a layer to be applied')
@click.option('--list', '-l', help='list of name of layers to applied, \
                          name of layers should be comma separated')
@click.argument('work_number')
@click.argument('out', type=click.File('w'))
def layer(**kwargs):
    """
    Command to apply a single layer, multiple layers or all available layers (by default) and then export to markdown.\n
    Args:\n
        - WORK_NUMBER is the work number of the pecha, from which given layer will be applied\n
        - OUT is the filename to the write the result. Currently support only Markdown file.
    """
    work_id = get_pecha_id(kwargs['work_number'])
    opfpath = config["OP_DATA_PATH"]/work_id/f'{work_id}.opf'
    serializer = SerializeMd(opfpath)
    if kwargs['name']:
        serializer.apply_layer(kwargs['name'])
    elif kwargs['list']:
        layers = kwargs['list'].split(',')
        serializer.layers = layers
        serializer.apply_layers()
    else:
        serializer.apply_layers()

    result = serializer.get_result()
    click.echo(result, file=kwargs['out'])

    # logging
    msg = f'Output is save at: {kwargs["out"].name}'
    click.echo(INFO.format(msg))


def pecha_list():
    return [pecha.name for pecha in config['OP_DATA_PATH'].iterdir()]

def get_data_path():
    return Path(config['DATA_CONFIG_PATH'].read_text().strip())


def check_edits(w_id):
    edit_path = get_data_path()
    data_path = config['OP_DATA_PATH']

    srcbl = (data_path/f'{w_id}'/f'{w_id}.opf'/'base.txt').read_text()
    dstbl = (edit_path/f'{w_id}.txt').read_text()

    return srcbl != dstbl, srcbl, dstbl


def setup_credential(repo):
    # setup authentication, if not done
    if not (config['CONFIG_PATH']/'credential').is_file():
        username = click.prompt('Github Username')
        password = click.prompt('Github Password', hide_input=True)
        # save credential
        (config['CONFIG_PATH']/'credential').write_text(f'{username},{password}')

    if not '@' in repo.remotes.origin.url:
        # get user credentials
        credential = (config['CONFIG_PATH']/'credential').read_text()
        username, password = [s.strip() for s in credential.split(',')]
        
        old_url = repo.remotes.origin.url.split('//')
        repo.remotes.origin.set_url(
            f'{old_url[0]}//{username}:{password}@{old_url[1]}'
        )
    
    return repo


def github_push(repo, branch_name, msg='made edits'):
    # credential
    repo = setup_credential(repo)

    # checkout to edited branch
    if branch_name in repo.heads:
        current = repo.heads[branch_name]
    else:
        current = repo.create_head(branch_name)
    current.checkout()

    # Add, commit and push the edited branch
    if repo.is_dirty():
        repo.git.add(A=True)
        repo.git.commit(m=msg)
        try: 
            repo.git.push('--set-upstream', 'origin', current)
        except Exception as e:
            print(e)
            msg = f'Authentication failed: Try again later'
            click.echo(ERROR.format(msg))
            return False

    # finally checkout to master for apply layer on validated text
    repo.heads['master'].checkout()
    
    return True


def repo_reset(repo, branch_name):
    # remove edited branch
    repo.heads['master'].checkout()
    repo.delete_head(repo.heads[branch_name], force=True)

    # reset to the origin url
    url = repo.remotes.origin.url.split('@')
    protocol = url[0].split('//')[0]
    repo.remotes.origin.set_url(
            f'{protocol}//{url[1]}'
    )


# Update annotations command
@cli.command()
@click.argument('work_number')
def update(**kwargs):
    """
    Command to update the base text with your edits.
    """
    work_id = get_pecha_id(kwargs['work_number'])
    if work_id:
        if work_id in pecha_list():
            repo_path = config["OP_DATA_PATH"]/work_id
            repo = Repo(str(repo_path))

            # if edited branch exists, then to check for changes in edited branch
            branch_name = 'edited'
            if branch_name in repo.heads:
                current = repo.heads[branch_name]
                current.checkout()

            is_changed, srcbl, dstbl = check_edits(work_id)
            if is_changed:
                msg = f'Updating {work_id} base text.'
                click.echo(INFO.format(msg))

                # Update layer annotations
                updater =  Blupdate(srcbl, dstbl)
                opfpath = repo_path/f'{work_id}.opf'
                updater.update_annotations(opfpath)

                # Update base-text
                src = get_data_path()/f'{work_id}.txt'
                dst = opfpath/'base.txt'
                shutil.copy(str(src), str(dst))

                # Create edited branch and push to Github
                status = github_push(repo, branch_name)

                # logging
                if status:
                    msg = f'Pecha edits {work_id} are uploaded for futher validation'
                    click.echo(INFO.format(msg))
                else:
                    repo_reset(repo, branch_name)
            else:
                msg = f'There are no changes in Pecha {work_id}'
                click.echo(ERROR.format(msg))
        else:
            msg = f'{work_id} does not exits, check the work-id'
            click.echo(ERROR.format(msg))


# OpenPecha Formatter
formatter_types = ['ocr', 'hfml(default)', 'tsadra']

@cli.command()
@click.option('--name', '-n', type=click.Choice(formatter_types),
                              help='Type of formatter')
@click.option('--id', '-i', type=int,
                            help='Id of the pecha')
@click.argument('input_path')
def format(**kwargs):
    '''
    Command to format pecha into opf
    '''
    formatter = HFMLFormatter()
    if kwargs['name'] == 'ocr':
        formatter = GoogleOCRFormatter()
    elif kwargs['name'] == 'tsadra':
        formatter = HFMLFormatter()

    formatter.create_opf(kwargs['input_path'], kwargs['id'])


@cli.command()
@click.option('--text_id', '-ti', type=str, help='text id of text')
@click.argument('pecha_num')
def edit(**kwargs):
    '''
    Command to export Pecha for editing work
    '''
    pecha_id = get_pecha_id(kwargs['pecha_num'])
    opf_path = f'{config["OP_DATA_PATH"]}/{pecha_id}/{pecha_id}.opf'

    serializer = SerializeHFML(opf_path, text_id=kwargs['text_id'])
    serializer.apply_layers()
    
    out_fn = f'{pecha_id}.txt'
    result = serializer.get_result()
    click.echo(result, file=open(out_fn, 'w'))

    # logging
    msg = f'Output is save at: {out_fn}'
    click.echo(INFO.format(msg))


'''catalog module contains all the functionalities necessary for managin
the catalog. Functonalities includes:
    - Creating opfs from input text
    - Assiging ID to the new opf
    - Updating the catalog with new opfs

'''

import yaml
import requests

from openpecha.formatters import *
from openpecha.github_utils import create_file, create_readme
from openpecha.github_utils import github_publish
from openpecha.utils import *


buildin_pipes = {
    'input': {
        'ocr_result_input': ocr_result_input
    },
    'release': {
        'create_release_with_assets': create_release_with_assets
    }
}


class CatalogManager:
    '''Manages the catalog'''

    def __init__(self, pipes=None, formatter_type=None, not_include_files=['releases'], last_id_fn='last_id'):
        self.repo_name = "openpecha-catalog"
        self.batch_path = "data/batch.csv"
        self.last_id_path = f"data/{last_id_fn}"
        self.batch = []
        self.last_id = self._get_last_id()
        self.FormatterClass = self._get_formatter_class(formatter_type)
        self.not_include_files = not_include_files
        self.pipes = pipes if pipes else buildin_pipes


    def _get_formatter_class(self, formatter_type):
        '''Returns formatter class based on the formatter-type'''
        if formatter_type == 'ocr':
            return GoogleOCRFormatter
        elif formatter_type == 'tsadra':
            return TsadraFormatter

    def _get_last_id(self):
        '''returns the id assigin to last opf pecha'''
        last_id_url = f'https://raw.githubusercontent.com/OpenPecha/openpecha-catalog/master/{self.last_id_path}'
        r = requests.get(last_id_url)
        return int(r.content.decode('utf-8').strip()[1:])

    
    def _add_id_url(self, row):
        id = row[0]
        row[0] = f'[{id}](https://github.com/OpenPecha/{id})'
        return row


    def update_catalog(self):
        '''Updates the catalog csv to have new opf-pechas metadata'''
        # update last_id
        content = self.batch[-1][0].strip()
        create_file(self.repo_name, self.last_id_path, content, "update last id of Pecha", update=True)

        # update last_id
        self.last_id = int(content[1:])

        # create batch.csv
        content = '\n'.join([','.join(row) for row in map(self._add_id_url, self.batch)]) + '\n'
        create_file(self.repo_name, self.batch_path, content, "create new batch")
        print('[INFO] Updated the OpenPecha catalog')

        # reset the batch
        self.batch = []


    def _get_catalog_metadata(self, pecha_path):
        meta_fn = pecha_path/f'{pecha_path.name}.opf/meta.yml'
        metadata = yaml.safe_load(meta_fn.open())
        catalog_metadata = [
            metadata['id'].split(':')[-1],
            metadata['source_metadata']['title'],
            metadata['source_metadata']['volume'],
            metadata['source_metadata']['author'],
            metadata['source_metadata']['id']
        ]
        self.batch.append(catalog_metadata)
        create_readme(metadata['source_metadata'], pecha_path)


    def format_and_publish(self, path):
        '''Convert input pecha to opf-pecha with id assigined'''
        formatter = self.FormatterClass()
        self.last_id += 1
        pecha_path = formatter.create_opf(path, self.last_id)
        self._get_catalog_metadata(pecha_path)
        github_publish(pecha_path, not_includes=self.not_include_files)
        return pecha_path


    def ocr_to_opf(self, path):
        self._process(
            path,
            'ocr_result_input',
            'create_release_with_assets'
        )


    def _process(self, path, input_method, release_method):
        print('[INFO] Getting input')
        raw_pecha_path = self.pipes['input'][input_method](path)
        print('[INFO] Convert Pecha to OPF')
        opf_pecha_path = self.format_and_publish(raw_pecha_path)
        print('[INFO] Release OPF pecha')
        self.pipes['release'][release_method](opf_pecha_path)


if __name__ == "__main__":
    catalog = CatalogManager(formatter_type='ocr', last_id_fn='ocr-machine-08_last_id')
    catalog.ocr_to_opf('./tests/data/formatter/google_ocr/W3CN472')
    catalog.update_catalog()


import re

from openpecha.serializers import Serialize

class SerializeFootNote(Serialize):
    """
    Serializes the foot-note for given text_id
    """

    def apply_annotation(self, vol_id, ann):
        payload = ''
        if ann['type'] == 'pagination':
            payload = f'[{ann["page_index"]}] {ann["page_info"]}\n'
            start_cc, end_cc = self._get_adapted_span(ann['span'], vol_id)
            self.add_chars(vol_id, start_cc, True, payload)
        elif 'peydurma' in ann['type']:
            payload = f'<{ann["note"]}>'
            start_cc, end_cc = self._get_adapted_span(ann['span'], vol_id)
            self.add_chars(vol_id, end_cc, False, payload)
        elif ann['type'] == 'correction':
            start_cc, end_cc = self._get_adapted_span(ann['span'], vol_id)
            error = self.base_layers[vol_id][start_cc: end_cc+1]
            self.base_layers[vol_id] = self.base_layers[vol_id][:start_cc] + ann['correction'] + \
                self.base_layers[vol_id][end_cc+1:]

            payload = f'<{error} སྡེ་དགེ།>'
            n_diff = len(ann['correction']) - len(error)
            if not n_diff == 0:
                self.n_char_shifted.append((ann['span']['start'], n_diff))
                start_cc, end_cc = self._get_adapted_span(ann['span'], vol_id)
            self.add_chars(vol_id, end_cc, False, payload)
            


    def generate_foot_notes(self, text):

        annotated_text = ''
        foot_notes = []
        notes = re.finditer(r'\<.*?\>', text)
        last_note_end_idx = 0
        for i, note in enumerate(notes):
            start, end = note.span(0)
            note_text = text[start: end]
            annotated_text += text[last_note_end_idx: start]
            foot_note_id = f'[^{i+1}K]'
            annotated_text += foot_note_id
            foot_notes.append(f'{foot_note_id}: {note_text[1:-1]}')
            last_note_end_idx = note.span(0)[1]
        else:
            annotated_text += text[last_note_end_idx:]

        return annotated_text + '\n\n' + '\n'.join(foot_notes)


    def get_result(self):
        self.layers.pop()
        annotated_text, text_id = Serialize.get_result(self)
        return self.generate_foot_notes(annotated_text), text_id

if __name__ == "__main__":
    OPF_PECHA_PATH = '../openpecha-user/.openpecha/data/P000002/P000002.opf'

    serializer = SerializeFootNote(OPF_PECHA_PATH, text_id='D1794', layers=['correction', 'peydurma-note', 'pagination'])
    serializer.apply_layers()
    annotated_text, text_id = serializer.get_result()
    print(annotated_text)


'''
This module contains format variable for all the annotations
'''

from collections import namedtuple


__all__ = [
    'Layer', 'Page', 'Text', 'Correction', 'ErrorCandidate', 'CrossVolSpan',
    'SubText', 'Peydurma', 'Chapter', 'Tsawa', 'Quotation', 'Sabche', 'Yigchung',
    'Book_title', 'Author']

# General layer format
Layer = {
    'id': None,              # Uique id for annotation of specific Pecha or Abstract work. type: str
    'annotation_type': None, # Name of annotation, type: str
    'revision': None,             # Revision number. type: int
    'annotations': []            # Annotations are stored in list . type: list
}

# Span of the annotation
Span = {
    'start': None,     # String index of starting character of the the annotation. type: int
    'end': None,       # String index of the ending character of the annotation. type: int 
}

# Page annotation
Page = {
    'id': None,             # Unique id for particular instance of annotation of a layer. type: str
    'span': Span,           # Span of particualr annotation, type: Span
    'page_index': None,     # Page number based on Volume specified in `part_of`. type: int
    'page_info': None,      # Page information. type: str
    'reference': None,      # Any reference of page, eg: img_url. type: str
}

# Cross vol Span
CrossVolSpan = {
    'vol': None,
    'span': Span
}

# Sub_text annotation
SubText = {
    'id': None,              # Unique id for the sub_text
    'work': None,            # index of the sub_text
    'span': []               # span of the sub_text
}

# Text annotation
Text = {
    'id': None,              # Unique id for particular Text. type: str
    'work': None,            # id of the `text`. type: int
    'parts': [],             # list of SubText
    'span': []               # list of CrossVolSpan
}

# Error and Suggestion annotation
Correction = {
    'id': None,             # Unque id for particular error text. type: int
    'correction': None,     # Correct text suggestion. type: str
    'certainty': None,      # Certainty of the suggested correct text. type: int
    'span': Span            # Span of error text. type: Span
}

ErrorCandidate = {
    'id': None,
    'span': Span
}

# Note
Peydurma = {
    'id': None,
    'note': None,
    'span': Span
}

# Book titlle and author
Book_title = {'span': Span}
Author = {'span': Span}

# Chapter
Chapter = {
    'id': None,
    'span': Span
}

# Tsawa
Tsawa = {
    'id': None,
    'span': Span
}

# Quotation
Quotation = {
    'id': None,
    'span': Span
}

# Sabche
Sabche = {
    'id': None,
    'span': Span
}

# Yigchung:
Yigchung = {
    'id': None,
    'span': Span
}


from pathlib import Path
from setuptools import setup, find_packages


def read(fname):
    p = Path(__file__).parent / fname
    with p.open(encoding="utf-8") as f:
        return f.read()

setup(
    name="openpecha",
    version="0.5.1",
    author="Esukhia developers",
    author_email="esukhiadev@gmail.com",
    description="OpenPecha Toolkit allows state of the art for distributed standoff annotations on moving texts",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="Apache2",
    url="https://github.com/OpenPoti/openpecha-toolkit",
    packages=find_packages(),
    install_requires=[
        'Click==7.0',
        'diff-match-patch==20181111',
        'PyYAML==5.1.2',
        'requests==2.22.0',
        'tqdm==4.35.0',
        'PyGithub==1.43.8',
        'GitPython==3.1.0',
        'bs4',
        'pyewts'
    ],
    python_requires=">=3.6",
    tests_require=["pytest"],
    entry_points={
        "console_scripts": ["openpecha=openpecha.cli:cli"] # command=package.module:function
    },
)



from pathlib import Path
from uuid import uuid4
import yaml


class BaseFormatter:
    '''
    OpenPecha Base class Formatter to parse annotated text into openpecha format.

    Example of OpenPoti format
    ==========================

        W1OP000001.opf
            ├── base.txt          # plain text, without markups (annotations)
            ├── layers            # layers with annotation's char coordinate
            │   ├── title.yml     
            │   ├── yigchung.yml
            |   ├── citation.yml
    '''

    def __init__(self, output_path='./output'):
        self.output_path = Path(output_path)


    def get_unique_id(self):
        return uuid4().hex


    def _build_dirs(self, input_path, id=None):
        '''
        Build the necessary directories for OpenPecha format.
        '''
        if id:
            pecha_id = f'P{id:06}'
        else:
            pecha_id = input_path.stem

        self.pecha_id = pecha_id
        self.dirs = {'opf_path': self.output_path/f'{pecha_id}/{pecha_id}.opf'}
        self.dirs['layers_path'] = self.dirs['opf_path']/'layers'
        self.dirs['base_path'] = self.dirs['opf_path']/'base'
        
        self.dirs['layers_path'].mkdir(parents=True, exist_ok=True)
        self.dirs['base_path'].mkdir(parents=True, exist_ok=True)

    
    def normalizeUni(self, strNFC):
        strNFC = strNFC.replace("\u0F00", "\u0F68\u0F7C\u0F7E") # ༀ
        strNFC = strNFC.replace("\u0F43", "\u0F42\u0FB7") # གྷ
        strNFC = strNFC.replace("\u0F48", "\u0F47\u0FB7") # ཈
        strNFC = strNFC.replace("\u0F4D", "\u0F4C\u0FB7") # ཌྷ
        strNFC = strNFC.replace("\u0F52", "\u0F51\u0FB7") # དྷ
        strNFC = strNFC.replace("\u0F57", "\u0F56\u0FB7") # བྷ
        strNFC = strNFC.replace("\u0F5C", "\u0F5B\u0FB7") # ཛྷ
        strNFC = strNFC.replace("\u0F69", "\u0F40\u0FB5") # ཀྵ
        strNFC = strNFC.replace("\u0F73", "\u0F71\u0F72") # ཱི
        strNFC = strNFC.replace("\u0F75", "\u0F71\u0F74") #  ཱུ
        strNFC = strNFC.replace("\u0F76", "\u0FB2\u0F80") # ྲྀ
        strNFC = strNFC.replace("\u0F77", "\u0FB2\u0F71\u0F80") # ཷ
        strNFC = strNFC.replace("\u0F78", "\u0FB3\u0F80") # ླྀ
        strNFC = strNFC.replace("\u0F79", "\u0FB3\u0F71\u0F80") # ཹ
        strNFC = strNFC.replace("\u0F81", "\u0F71\u0F80") #  ཱྀ
        strNFC = strNFC.replace("\u0F93", "\u0F92\u0FB7") # ྒྷ
        strNFC = strNFC.replace("\u0F9D", "\u0F9C\u0FB7") # ྜྷ
        strNFC = strNFC.replace("\u0FA2", "\u0FA1\u0FB7") # ྡྷ
        strNFC = strNFC.replace("\u0FA7", "\u0FA6\u0FB7") # ྦྷ
        strNFC = strNFC.replace("\u0FAC", "\u0FAB\u0FB7") # ྫྷ
        strNFC = strNFC.replace("\u0FB9", "\u0F90\u0FB5") # ྐྵ
        return strNFC


    def text_preprocess(self, text):
        raise NotImplementedError('Text preprocessing depends on type of text format, \
                                   should be implemented in sub-class.')


    def get_input(self, input_path):
        '''
        Return a preprocess text from given input_file path
        '''
        m_text = self.text_preprocess(input_path.read_text())
        return self.normalizeUni(m_text)


    def format_layer(self, layers):
        '''
        Post-processing for various layer to easily dump into yaml file. For eg, title annotation has only one char coord, #<title_text>
        yigchung annotations has part of char coord, *<yigchun_text>*.
        '''
        raise NotImplementedError('Layers PostProcessing depends on the type of annotations, \
                                   should be implemented in sub-class.')


    def build_layers(self, text):
        '''
        Parse all the layers annotation from the given text.
        '''
        raise NotImplementedError('Parsing annotation depends type of annotation in the text, \
                                  should be implemented in sub-class.')


    def get_base_text(self, m_text):
        'Retuns text with all annotation removed'
        raise NotImplementedError('Every type of text have different format for annotation, \
                                  should be implemented in sub_class.')

    
    def dump(self, data, output_fn):
        with output_fn.open('w') as fn:
            yaml.dump(data, fn, default_flow_style=False,  sort_keys=False, allow_unicode=True)
        
    
    def load(self, fn):
        return yaml.safe_load(fn.open())


    def create_opf(self, input_path):
        input_path = Path(input_path)
        self._build_dirs(input_path)

        m_text = self.get_input(input_path)
        layers = self.build_layers(m_text)
        base_text = self.get_base_text(m_text)

        # save layers
        for layer, ann in layers.items():
            layer_fn = self.dirs['layers_path']/f'{layer}.yml'
            self.dump(ann, layer_fn)

        # save base_text
        (self.dirs['opf_path']/'base.txt').write_text(base_text)